#ifndef __DH_H__
#define __DH_H__

/**
 * Functions useful for calculating the forward kinematics of a robot.
 * Some is taken from the Motion Strategy Library and some is taken from 
 * the book by Spong, Hutchinson & Vidyasager p. 77. 
 */

#include <MSL/vector.h>
#include <MSL/matrix.h>

#include <math.h>

#ifndef sqr
#define sqr(x) (x*x)
#endif

#ifndef PI
#define PI 3.14159265
#endif

class DH
{
 public:
  /**
   * Calculates transformation matrix from Denavit-Hartenberg parameters.
   * Taken from Spong, Hutchinson & Vidyasager p. 77. 
   */
  static MSLMatrix T(double A, double Alpha, double D, double Theta) {
    MSLMatrix r(4,4);
    r[0][0] = cos(Theta);
    r[1][0] = sin(Theta);
    r[2][0] = 0;
    r[3][0] = 0;
    r[0][1] = -sin(Theta) * cos(Alpha);
    r[1][1] = cos(Theta) * cos(Alpha);
    r[2][1] = sin(Alpha);
    r[3][1] = 0;
    r[0][2] = sin(Theta)*sin(Alpha);
    r[1][2] = -cos(Theta)*sin(Alpha);
    r[2][2] = cos(Alpha);
    r[3][2] = 0;
    r[0][3] = A * cos(Theta);
    r[1][3] = A * sin(Theta);
    r[2][3] = D;
    r[3][3] = 1.0;
    return r;
  }

  static MSLMatrix inverse(const MSLMatrix &m) {
    MSLMatrix R(3,3);
    R[0][0] = m[0][0];
    R[0][1] = m[0][1];
    R[0][2] = m[0][2];
    R[1][0] = m[1][0];
    R[1][1] = m[1][1];
    R[1][2] = m[1][2];
    R[2][0] = m[2][0];
    R[2][1] = m[2][1];
    R[2][2] = m[2][2];
    MSLVector d(3);
    d[0] = m[0][3];
    d[1] = m[1][3];
    d[2] = m[2][3];
    MSLMatrix result(4,4);
    R = R.trans();
    d = -R*d;
    result[0][0] = R[0][0];
    result[0][1] = R[0][1];
    result[0][2] = R[0][2];
    result[1][0] = R[1][0];
    result[1][1] = R[1][1];
    result[1][2] = R[1][2];
    result[2][0] = R[2][0];
    result[2][1] = R[2][1];
    result[2][2] = R[2][2];
    result[0][3] = d[0];
    result[1][3] = d[1];
    result[2][3] = d[2];
    result[3][3] = 1.0;
    result[3][0] = 0.0f;
    result[3][1] = 0.0f;
    result[3][2] = 0.0f;
    return result;
  }

  /**
   * Multiply two matrices and store the result in r
   * r = ro * rn
   */
  static void mul(MSLMatrix &r, MSLMatrix &ro, MSLMatrix &rn) {
    for (int k=0; k<4; k++){
      for (int l=0; l<4; l++) {
	r[k][l]=0.0 ;
	for (int m=0; m<4; m++){
	  r[k][l]+=ro[k][m]*rn[m][l];
	}
      }
    }
  }

  /**
   * Extract position and orientation from a transformation matrix r.
   * The information is stored in the vector q at offset offset.
   */
  static void extract(MSLVector &q, unsigned int offset, MSLMatrix &r) {
    q[offset] = r[0][3];
    q[offset+1] = r[1][3];
    q[offset+2] = r[2][3];


    q[offset+4] = atan2(-r[2][0], sqrt(sqr(r[0][0])+sqr(r[1][0])));
    if ((q[offset+4] > 0) && (cos(q[offset+4])==0.0)) {
      q[offset+5] = atan2(r[1][0], r[0][0]);
      q[offset+3] = 0.0f;
    } else if ((q[offset+4] < 0) && (cos(q[offset+4])==0.0)) {
      q[offset+5] = -atan2(r[1][0], r[0][0]);
      q[offset+3] = 0.0f;
    } else {
      q[offset+5] = atan2(r[1][0], r[0][0]);
      q[offset+3] = atan2(r[2][1], r[2][2]);
    }

    if (q[offset+3] < -PI)
      q[offset+3] += 2.0*PI; // Force the orientation into [-pi,pi)
    else if (q[offset+3] >= PI)
      q[offset+3] -= 2.0*PI;
    if (q[offset+4] < -PI)
      q[offset+4] += 2.0*PI; // Force the orientation into [-pi,pi)
    else if (q[offset+4] >= PI)
      q[offset+4] -= 2.0*PI;
    if (q[offset+5] < -PI)
      q[offset+5] += 2.0*PI; // Force the orientation into [-pi,pi)
    else if (q[offset+5] >= PI)
      q[offset+5] -= 2.0*PI;
  }

  static void identity(MSLMatrix &r) {
    for (int i = 0; i < 4 ; i ++ ) {
      for (int j = 0 ; j < 4 ; j ++ ) {
	if (i==j) {r[i][j] = 1.0;}
	else {r[i][j] = 0.0;}
      }
    }
  }

  static void zeros(MSLMatrix &r) {
    for (int i = 0; i < 4 ; i ++ ) {
      for (int j = 0 ; j < 4 ; j ++ ) {
	r[i][j] = 0.0;
      }
    }
  }
};

#endif
