#include <QWidget>
#include <GL/gl.h>

#include "irbrenderer.h"
#include "transform.h"

IRBRenderer::IRBRenderer(string path, ModelIRB *model) : QObject(), path(path), model(model)
{
  float colors[6][3] = {{1.0, 0.0, 0.0},
			{0.0, 1.0, 0.0},
			{0.0, 0.0, 1.0},
			{1.0, 1.0, 0.0},
			{0.0, 1.0, 1.0},
			{1.0, 0.0, 1.0}};

  Link *collisionLink, *graphicalLink;
  Frame *frame;

  for (int i=0; i<model->NumBodies; i++) {
    std::stringstream ss;
    ss << i;
    collisionLink = new Link(path + string("/Robot")+ss.str());
    graphicalLink = new Link(path + string("/graphics/Robot")+ss.str());
    frame = new Frame;

    collisionLink->setColor(colors[i%6][0], colors[i%6][1], colors[i%6][2]);
    graphicalLink->setColor(colors[i%6][0], colors[i%6][1], colors[i%6][2]);

    collisionLinks.push_back(collisionLink);
    graphicalLinks.push_back(graphicalLink);
    frames.push_back(frame);
  }
}

QWidget* IRBRenderer::getControlWidget(unsigned int index)
{
  if (index >= collisionLinks.size()) {
    return NULL;
  }
  
  Link *cLink = collisionLinks[index];
  Link *gLink = graphicalLinks[index];
  
  Transform *cTrans = new Transform(string("Collision link"), cLink);
  Transform *gTrans = new Transform(string("Graphical link"), gLink);
  QObject::connect(cTrans, SIGNAL(update()),
		   this, SLOT(refresh()));
  QObject::connect(gTrans, SIGNAL(update()),
		   this, SLOT(refresh()));

  QWidget *widget = new QWidget;
  QVBoxLayout *layout = new QVBoxLayout;

  widget->setLayout(layout);
  layout->addWidget(cTrans);
  layout->addWidget(gTrans);

  return widget;
}

string IRBRenderer::getName(unsigned int index)
{
  std::stringstream ss;
  ss << index;
  string r;
  ss >> r;
  return string("Link ") + r;
}

void IRBRenderer::renderGraphicalModels()
{
  for (unsigned int i=0; i<graphicalLinks.size(); i++) {
    graphicalLinks[i]->draw();
  }
}

void IRBRenderer::renderCollisionModels()
{
  for (unsigned int i=0; i<collisionLinks.size(); i++) {
    collisionLinks[i]->draw();
  }
}

void IRBRenderer::renderCoordinateSystems()
{
  for (unsigned int i=0; i<frames.size(); i++) {
    frames[i]->drawCoordinateSystem();
  }
}

void IRBRenderer::renderLineModel()
{
  glBegin(GL_LINES);
  glVertex3f(1.0, 1.0, 1.0);
  for (unsigned int i=0; i<frames.size()-1; i++) {
    glVertex3f(frames[i]->getPosX(), frames[i]->getPosY(), frames[i]->getPosZ());
    glVertex3f(frames[i+1]->getPosX(), frames[i+1]->getPosY(), frames[i+1]->getPosZ());
  }
  glEnd();
}

void IRBRenderer::stateToConfiguration(MSLVector state)
{
  MSLVector q;
  q = model->StateToConfiguration(state);

  for (int i=0; i<state.dim(); i++) {
    collisionLinks[i]->setTransform(q[i*6], q[i*6+1], q[i*6+2], q[i*6+3], q[i*6+4], q[i*6+5]);
    graphicalLinks[i]->setTransform(q[i*6], q[i*6+1], q[i*6+2], q[i*6+3], q[i*6+4], q[i*6+5]);
    frames[i]->setTransform(q[i*6], q[i*6+1], q[i*6+2], q[i*6+3], q[i*6+4], q[i*6+5]);    
  }
  emit update();
}

unsigned int IRBRenderer::numLinks()
{
  return collisionLinks.size();
}

void IRBRenderer::refresh()
{
  emit update();
}

#include "moc_irbrenderer.cpp"
