//#include <QGLWidget>
#include <GL/gl.h>
#include <fstream>
#include <math.h>

#include "link.h"

Link::Link(string filename) : Frame()
{
  this->filename = filename;
  string buffer;

  ifstream fin(filename.c_str());
  while(getline(fin, buffer)) {    
    polygons.push_back(Triangle(buffer));
  }
  fin.close();
}

void Link::draw()
{
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glTranslatef(pos[0], pos[1], pos[2]);

  glRotatef(rot[2]*180.0/PI, 0.0, 0.0, 1.0);
  glRotatef(rot[1]*180.0/PI, 0.0, 1.0, 0.0);
  glRotatef(rot[0]*180.0/PI, 1.0, 0.0, 0.0);

  glColor3f(color[0], color[1], color[2]);
  glBegin(GL_TRIANGLES);
  for (unsigned int i=0;i<polygons.size(); i++) {
    glNormal3f(polygons[i].nx,   polygons[i].ny,   polygons[i].nz);
    glVertex3f(polygons[i].pt1x, polygons[i].pt1y, polygons[i].pt1z);
    glNormal3f(polygons[i].nx,   polygons[i].ny,   polygons[i].nz);
    glVertex3f(polygons[i].pt2x, polygons[i].pt2y, polygons[i].pt2z);
    glNormal3f(polygons[i].nx,   polygons[i].ny,   polygons[i].nz);
    glVertex3f(polygons[i].pt3x, polygons[i].pt3y, polygons[i].pt3z);
  }
  glEnd();
  glPopMatrix();
}

void Link::translate(float x, float y, float z)
{
  for (unsigned int i=0; i<polygons.size(); i++) {
    polygons[i].pt1x += x;
    polygons[i].pt1y += y;
    polygons[i].pt1z += z;
    polygons[i].pt2x += x;
    polygons[i].pt2y += y;
    polygons[i].pt2z += z;
    polygons[i].pt3x += x;
    polygons[i].pt3y += y;
    polygons[i].pt3z += z;
    polygons[i].calcNormal();
  }
}

void Link::rotate(double psi_deg, double theta_deg, double phi_deg)
{
  double psi = psi_deg * PI / 180.0f;
  double theta = theta_deg * PI / 180.0f;
  double phi = phi_deg * PI / 180.0f;
  double r00 = cos(psi) * cos(theta);
  double r10 = sin(psi) * cos(theta);
  double r20 = -sin(theta);
  double r01 = -sin(psi) * cos(phi) + cos(psi) * sin(theta) * sin(phi);
  double r11 = cos(psi) * cos(phi) + sin(psi) * sin(theta) * sin(phi);
  double r21 = cos(theta)*sin(phi);
  double r02 = sin(psi)*sin(phi) + cos(psi)*sin(theta)*cos(phi);
  double r12 = -cos(psi)*sin(phi) + sin(psi)*sin(theta)*cos(phi);
  double r22 = cos(theta)*cos(phi);

  double nx, ny, nz;

  for (unsigned int i=0; i<polygons.size(); i++) {
    nx = r00 * polygons[i].pt1x + r01 * polygons[i].pt1y + r02 * polygons[i].pt1z;
    ny = r10 * polygons[i].pt1x + r11 * polygons[i].pt1y + r12 * polygons[i].pt1z;
    nz = r20 * polygons[i].pt1x + r21 * polygons[i].pt1y + r22 * polygons[i].pt1z;
    polygons[i].pt1x = nx;
    polygons[i].pt1y = ny;
    polygons[i].pt1z = nz;

    nx = r00 * polygons[i].pt2x + r01 * polygons[i].pt2y + r02 * polygons[i].pt2z;
    ny = r10 * polygons[i].pt2x + r11 * polygons[i].pt2y + r12 * polygons[i].pt2z;
    nz = r20 * polygons[i].pt2x + r21 * polygons[i].pt2y + r22 * polygons[i].pt2z;
    polygons[i].pt2x = nx;
    polygons[i].pt2y = ny;
    polygons[i].pt2z = nz;

    nx = r00 * polygons[i].pt3x + r01 * polygons[i].pt3y + r02 * polygons[i].pt3z;
    ny = r10 * polygons[i].pt3x + r11 * polygons[i].pt3y + r12 * polygons[i].pt3z;
    nz = r20 * polygons[i].pt3x + r21 * polygons[i].pt3y + r22 * polygons[i].pt3z;
    polygons[i].pt3x = nx;
    polygons[i].pt3y = ny;
    polygons[i].pt3z = nz;

    polygons[i].calcNormal();
  }
}

void Link::scale(double x, double y, double z)
{
  for (unsigned int i=0; i<polygons.size(); i++) {
    polygons[i].pt1x = polygons[i].pt1x*x;
    polygons[i].pt2x = polygons[i].pt2x*x;
    polygons[i].pt3x = polygons[i].pt3x*x;

    polygons[i].pt1y = polygons[i].pt1y*y;
    polygons[i].pt2y = polygons[i].pt2y*y;
    polygons[i].pt3y = polygons[i].pt3y*y;

    polygons[i].pt1z = polygons[i].pt1z*z;
    polygons[i].pt2z = polygons[i].pt2z*z;
    polygons[i].pt3z = polygons[i].pt3z*z;
  }
}

void Link::switchDirection()
{
  double tmp1, tmp2, tmp3;
  for (unsigned int i=0; i<polygons.size(); i++) {
    tmp1 = polygons[i].pt1x;
    tmp2 = polygons[i].pt1y;
    tmp3 = polygons[i].pt1z;
    polygons[i].pt1x = polygons[i].pt3x;
    polygons[i].pt1y = polygons[i].pt3y;
    polygons[i].pt1z = polygons[i].pt3z;
    polygons[i].pt3x = tmp1;
    polygons[i].pt3y = tmp2;
    polygons[i].pt3z = tmp3;
    polygons[i].calcNormal();
  }
}

void Link::save()
{
  ofstream fout(filename.c_str());
  for (unsigned int i=0; i<polygons.size(); i++) {
    fout << "(" << polygons[i].pt1x << ","<< polygons[i].pt1y << ","<< polygons[i].pt1z << ") ("
	 << polygons[i].pt2x << ","<< polygons[i].pt2y << ","<< polygons[i].pt2z << ") ("
	 << polygons[i].pt3x << ","<< polygons[i].pt3y << ","<< polygons[i].pt3z <<")"<<endl;
  }
  fout.close();
}
