#include <math.h>
#include <MSL/model.h>
#include <MSL/defs.h>
#include "modelirbsingle.h"
#include "dh.h"

#ifndef sqr
#define sqr(x) (x*x)
#endif
#ifndef PI
#define PI 3.14159265
#endif

ModelIRBSingle::ModelIRBSingle(string path) : ModelIRB(path)
{
  int i;
  MSLVector u;
  std::ifstream fin;

  READ_PARAMETER_OR_ERROR(NumBodies);
  READ_PARAMETER_OR_ERROR(StateDim);
  READ_PARAMETER_OR_ERROR(DH);

  InputDim = StateDim; 

  StateIndices = vector<int>(StateDim);
  fin.open((FilePath + "StateIndices").c_str());
  if (fin) {
    for (i = 0; i < StateDim; i++) {
      fin >> StateIndices[i];
    }
  }
  else {
    cout << "Error: No StateIndices file was found\n";
    exit(-1);
  }
  fin.close();

  LowerState = MSLVector(StateDim);
  UpperState = MSLVector(StateDim);

  for (i = 0; i < StateDim; i++)
    LowerState[i] = -1.0;
  for (i = 0; i < StateDim; i++) 
    UpperState[i] = 1.0;

  READ_OPTIONAL_PARAMETER(LowerState);
  READ_OPTIONAL_PARAMETER(UpperState);

  Inputs.clear();  // Otherwise its parent constructor will make some inputs
  for (i = 0; i < StateDim; i++) {
    u = MSLVector(StateDim);
    u[i] = 0.05;
    Inputs.push_back(u);
    u = MSLVector(StateDim);
    u[i] = -0.05;
    Inputs.push_back(u);
  }    

  READ_OPTIONAL_PARAMETER(Inputs);

  LoadCornerPoints();
}

MSLVector ModelIRBSingle::StateToConfiguration(const MSLVector &x)
{
  // I didn't understand "their" DH convention
  // so I reimplemented it.
  MSLVector q;
  MSLVector A, Alpha, D, Theta;
  int i;
  MSLMatrix r(4,4), rn(4,4), ro(4,4);

  q = MSLVector(6*NumBodies);
  A = MSLVector(NumBodies);
  D = MSLVector(NumBodies);
  Alpha = MSLVector(NumBodies);
  Theta = MSLVector(NumBodies);

  for (i = 0; i < StateDim; i++ ) {
    if (StateIndices[i] != 0) {
      int y = StateIndices[i];
      DH[y-1] = x[i];
    }
  } 
 
  for (i = 0; i < NumBodies; i++) {
    Alpha[i] = DH[i];
    Theta[i] = DH[NumBodies*1+i];
    A[i] = DH[NumBodies*2+i];
    D[i] = DH[NumBodies*3+i];
  }

  DH::identity(r);
  
  for (i = 0; i < NumBodies; i++) {
    ro=r;
    rn = DH::T(A[i], Alpha[i], D[i], Theta[i]);
    DH::mul(r,ro,rn);
    DH::extract(q, 6*i, r);
  }

  angle(q);

  return q;  
}

float ModelIRBSingle::MaximumRobotDisplacement(const MSLVector &deltaq)
{
  // Array of Lipschitz constants
  float c[6];
  c[0] = 1;
  c[1] = 1;
  c[2] = 1;
  c[3] = 1;
  c[4] = 1;
  c[5] = 1;

  float res = 0.0f;
  for (int i=0; i<6; i++) {
    res = res + c[i] * fabs(deltaq[i]);
  }
  return res;
}

/**
 * Check whether the angle of the end effector's y-axis and the vertical is less than gamma.
 */
bool ModelIRBSingle::angle(const MSLVector &q)
{
  double psi = q[6*5+5];
  double theta = q[6*5+4];
  double phi = q[6*5+3];

  // Calculate y axis vector for end effector of robot 1
  MSLVector v1(cos(psi)*sin(theta)*sin(phi) - sin(psi)*cos(phi),
	       sin(psi)*sin(theta)*sin(phi) + cos(psi)*cos(phi), 
	       cos(theta)*sin(phi));

  MSLVector v2(0.0, 0.0, 1.0);
  double angle = v1.angle(v2);
  //  cout << "Angle: " << angle << endl;
  if (fabs(angle) < 0.2)
    return true;

  return false;
}

/**
 * Check that distance and angles between end effectors is within tolerances.
 */
bool ModelIRBSingle::Satisfied(const MSLVector &x)
{
  MSLVector q = StateToConfiguration(x);

  return angle(q);
}

void ModelIRBSingle::LoadCornerPoints()
{  
  LoadCornerPointsFromFile("data/irb/Robot0", CornerRobot0);
  LoadCornerPointsFromFile("data/irb/Robot1", CornerRobot1);
  LoadCornerPointsFromFile("data/irb/Robot2", CornerRobot2);
  LoadCornerPointsFromFile("data/irb/Robot3", CornerRobot3);
  LoadCornerPointsFromFile("data/irb/Robot4", CornerRobot4);
  LoadCornerPointsFromFile("data/irb/Robot5", CornerRobot5);
}

void ModelIRBSingle::LoadCornerPointsFromFile(string filename, vector<MSLVector> &corners)
{
  float pts[9];
  string buffer;
  ifstream fin(filename.c_str());

  while(getline(fin, buffer)) {
    sscanf(buffer.c_str(), "(%f,%f,%f) (%f,%f,%f) (%f,%f,%f)",
	   &pts[0], &pts[1], &pts[2],
	   &pts[3], &pts[4], &pts[5],
	   &pts[6], &pts[7], &pts[8]);

    AddCornerPoint(MSLVector(pts[0], pts[1], pts[2]), corners);
    AddCornerPoint(MSLVector(pts[3], pts[4], pts[5]), corners);
    AddCornerPoint(MSLVector(pts[6], pts[7], pts[8]), corners);
  }  
}

void ModelIRBSingle::AddCornerPoint(MSLVector point, vector<MSLVector> & corners)
{
  for (unsigned int i=0; i<corners.size(); i++) {
    MSLVector diff = corners[i] - point;
    if (diff.length() < 0.01)
      return;
  }
  corners.push_back(point);
}

void ModelIRBSingle::LipschitzConstants(const MSLVector &x, float c[36])
{
    MSLVector q;
    MSLVector A, Alpha, D, Theta;
    int i;
    MSLMatrix r(4,4), rn(4,4), ro(4,4);
    MSLMatrix r0(4,4), r1(4,4), r2(4,4), r3(4,4), r4(4,4), r5(4,4);

    q = MSLVector(6*NumBodies);
    A = MSLVector(NumBodies);
    D = MSLVector(NumBodies);
    Alpha = MSLVector(NumBodies);
    Theta = MSLVector(NumBodies);
    
    for (i = 0; i < StateDim; i++ ) {
      if (StateIndices[i] != 0) {
	int y = StateIndices[i];
	DH[y-1] = x[i];
      }
    } 
    
    for (i = 0; i < NumBodies; i++) {
      Alpha[i] = DH[i];
      Theta[i] = DH[NumBodies*1+i];
      A[i] = DH[NumBodies*2+i];
      D[i] = DH[NumBodies*3+i];
    }
    
    DH::identity(r);

    r0 = DH::T(A[0], Alpha[0], D[0], Theta[0]);
    MSLMatrix temp = DH::T(A[1], Alpha[1], D[1], Theta[1]);
    DH::mul(r1, r0, temp);
    temp = DH::T(A[2], Alpha[2], D[2], Theta[2]);
    DH::mul(r2, r1, temp);
    temp = DH::T(A[3], Alpha[3], D[3], Theta[3]);
    DH::mul(r3, r2, temp);
    temp = DH::T(A[4], Alpha[4], D[4], Theta[4]);
    DH::mul(r4, r3, temp);
    temp = DH::T(A[5], Alpha[5], D[5], Theta[5]);
    DH::mul(r5, r4, temp);
    
    MSLMatrix r1inverse = DH::inverse(r1);
    MSLMatrix r2inverse = DH::inverse(r2);
    MSLMatrix r3inverse = DH::inverse(r3);
    MSLMatrix r4inverse = DH::inverse(r4);
    MSLMatrix r5inverse = DH::inverse(r5);
    MSLVector z_0t0 = MSLVector(4);
    z_0t0[0] = r0[0][2];
    z_0t0[1] = r0[1][2];
    z_0t0[2] = r0[2][2];
    z_0t0[3] = 1.0f;
    MSLVector z_0t1 = r1inverse*z_0t0;
    MSLVector z_0t2 = r2inverse*z_0t0;
    MSLVector z_0t3 = r3inverse*z_0t0;
    MSLVector z_0t4 = r4inverse*z_0t0;
    MSLVector z_0t5 = r5inverse*z_0t0;
    MSLVector z_1t0 = MSLVector(4);
    z_1t0[0] = r1[0][2];
    z_1t0[1] = r1[1][2];
    z_1t0[2] = r1[2][2];
    z_1t0[3] = 1.0f;
    MSLVector z_1t1 = r1inverse*z_1t0;
    MSLVector z_1t2 = r2inverse*z_1t0;
    MSLVector z_1t3 = r3inverse*z_1t0;
    MSLVector z_1t4 = r4inverse*z_1t0;
    MSLVector z_1t5 = r5inverse*z_1t0;
    MSLVector z_2t0 = MSLVector(4);
    z_2t0[0] = r2[0][2];
    z_2t0[1] = r2[1][2];
    z_2t0[2] = r2[2][2];
    z_2t0[3] = 1.0f;
    MSLVector z_2t2 = r2inverse*z_2t0;
    MSLVector z_2t3 = r3inverse*z_2t0;
    MSLVector z_2t4 = r4inverse*z_2t0;
    MSLVector z_2t5 = r5inverse*z_2t0;
    MSLVector z_3t0 = MSLVector(4);
    z_3t0[0] = r3[0][2];
    z_3t0[1] = r3[1][2];
    z_3t0[2] = r3[2][2];
    z_3t0[3] = 1.0f;
    MSLVector z_3t3 = r3inverse*z_3t0;
    MSLVector z_3t4 = r4inverse*z_3t0;
    MSLVector z_3t5 = r5inverse*z_3t0;
    MSLVector z_4t0 = MSLVector(4);
    z_4t0[0] = r4[0][2];
    z_4t0[1] = r4[1][2];
    z_4t0[2] = r4[2][2];
    z_4t0[3] = 1.0f;
    MSLVector z_4t4 = r4inverse*z_4t0;
    MSLVector z_4t5 = r5inverse*z_4t0;

    MSLVector z_5t0 = MSLVector(4);
    z_5t0[0] = r5[0][2];
    z_5t0[1] = r5[1][2];
    z_5t0[2] = r5[2][2];
    z_5t0[3] = 1.0f;
    MSLVector z_5t5 = r5inverse*z_5t0;

    MSLVector o_0t0 = MSLVector(4);
    o_0t0[0] = r0[0][3];
    o_0t0[1] = r0[1][3];
    o_0t0[2] = r0[2][3];
    o_0t0[3] = 1.0f;
    MSLVector o_0t1 = r1inverse*o_0t0;
    MSLVector o_0t2 = r2inverse*o_0t0;
    MSLVector o_0t3 = r3inverse*o_0t0;
    MSLVector o_0t4 = r4inverse*o_0t0;
    MSLVector o_0t5 = r5inverse*o_0t0;

    MSLVector o_1t0 = MSLVector(4);
    o_1t0[0] = r1[0][3];
    o_1t0[1] = r1[1][3];
    o_1t0[2] = r1[2][3];
    o_1t0[3] = 1.0f;
    MSLVector o_1t1 = r1inverse*o_1t0;
    MSLVector o_1t2 = r2inverse*o_1t0;
    MSLVector o_1t3 = r3inverse*o_1t0;
    MSLVector o_1t4 = r4inverse*o_1t0;
    MSLVector o_1t5 = r5inverse*o_1t0;

    MSLVector o_2t0 = MSLVector(4);
    o_2t0[0] = r2[0][3];
    o_2t0[1] = r2[1][3];
    o_2t0[2] = r2[2][3];
    o_2t0[3] = 1.0f;
    MSLVector o_2t2 = r2inverse*o_2t0;
    MSLVector o_2t3 = r3inverse*o_2t0;
    MSLVector o_2t4 = r4inverse*o_2t0;
    MSLVector o_2t5 = r5inverse*o_2t0;

    MSLVector o_3t0 = MSLVector(4);
    o_3t0[0] = r3[0][3];
    o_3t0[1] = r3[1][3];
    o_3t0[2] = r3[2][3];
    o_3t0[3] = 1.0f;
    MSLVector o_3t3 = r3inverse*o_3t0;
    MSLVector o_3t4 = r4inverse*o_3t0;
    MSLVector o_3t5 = r5inverse*o_3t0;

    MSLVector o_4t0 = MSLVector(4);
    o_4t0[0] = r4[0][3];
    o_4t0[1] = r4[1][3];
    o_4t0[2] = r4[2][3];
    o_4t0[3] = 1.0f;
    MSLVector o_4t4 = r4inverse*o_4t0;
    MSLVector o_4t5 = r5inverse*o_4t0;

    MSLVector o_5t0 = MSLVector(4);
    o_5t0[0] = r5[0][3];
    o_5t0[1] = r5[1][3];
    o_5t0[2] = r5[2][3];
    o_5t0[3] = 1.0f;
    MSLVector o_5t5 = r5inverse*o_5t0;


    for (unsigned int i=0; i<36; i++) {
      c[i] = 0.0f;
    }

    double dist;
    for (unsigned int i=0; i<CornerRobot0.size(); i++) {
      dist = PointToLineDistance(CornerRobot0[i], o_0t0, z_0t0);
      if (dist > c[0])
	c[0] = dist;      
    }

    for (unsigned int i=0; i<CornerRobot1.size(); i++) {
      dist = PointToLineDistance(CornerRobot1[i], o_0t1, z_0t1);
      if (dist > c[1])
	c[1] = dist;      
      dist = PointToLineDistance(CornerRobot1[i], o_1t1, z_1t1);
      if (dist > c[1+6])
	c[1+6] = dist;      
    }

    for (unsigned int i=0; i<CornerRobot2.size(); i++) {
      dist = PointToLineDistance(CornerRobot2[i], o_0t2, z_0t2);
      if (dist > c[2])
	c[2] = dist;      
      dist = PointToLineDistance(CornerRobot2[i], o_1t2, z_1t2);
      if (dist > c[2+6])
	c[2+6] = dist;      
      dist = PointToLineDistance(CornerRobot2[i], o_2t2, z_2t2);
      if (dist > c[2+12])
	c[2+12] = dist;      
    }

    for (unsigned int i=0; i<CornerRobot3.size(); i++) {
      dist = PointToLineDistance(CornerRobot3[i], o_0t3, z_0t3);
      if (dist > c[3])
	c[3] = dist;   
      dist = PointToLineDistance(CornerRobot3[i], o_1t3, z_1t3);
      if (dist > c[3+6])
	c[3+6] = dist;      
      dist = PointToLineDistance(CornerRobot3[i], o_2t3, z_2t3);
      if (dist > c[3+12])
	c[3+12] = dist;      
      dist = PointToLineDistance(CornerRobot3[i], o_3t3, z_3t3);
      if (dist > c[3+18])
	c[3+18] = dist;      
    }

    for (unsigned int i=0; i<CornerRobot4.size(); i++) {
      dist = PointToLineDistance(CornerRobot4[i], o_0t4, z_0t4);
      if (dist > c[4])
	c[4] = dist;   
      dist = PointToLineDistance(CornerRobot4[i], o_1t4, z_1t4);
      if (dist > c[4+6])
	c[4+6] = dist;      
      dist = PointToLineDistance(CornerRobot4[i], o_2t4, z_2t4);
      if (dist > c[4+12])
	c[4+12] = dist;      
      dist = PointToLineDistance(CornerRobot4[i], o_3t4, z_3t4);
      if (dist > c[4+18])
	c[4+18] = dist;      
      dist = PointToLineDistance(CornerRobot4[i], o_4t4, z_4t4);
      if (dist > c[4+24])
	c[4+24] = dist;      
    }

    for (unsigned int i=0; i<CornerRobot5.size(); i++) {
      dist = PointToLineDistance(CornerRobot5[i], o_0t5, z_0t5);
      if (dist > c[5])
	c[5] = dist;   
      dist = PointToLineDistance(CornerRobot5[i], o_1t5, z_1t5);
      if (dist > c[5+6])
	c[5+6] = dist;      
      dist = PointToLineDistance(CornerRobot5[i], o_2t5, z_2t5);
      if (dist > c[5+12])
	c[5+12] = dist;      
      dist = PointToLineDistance(CornerRobot5[i], o_3t5, z_3t5);
      if (dist > c[5+18])
	c[5+18] = dist;      
      dist = PointToLineDistance(CornerRobot5[i], o_4t5, z_4t5);
      if (dist > c[5+24])
	c[5+24] = dist;      
      dist = PointToLineDistance(CornerRobot5[i], o_5t5, z_5t5);
      if (dist > c[5+30])
	c[5+30] = dist;      
    }

    /*
    for (unsigned int i=0; i<6; i++) {
      cout << "c";
      for (unsigned int j=0; j<6; j++) {
	cout << "\t" << c[j+i*6];
      }
      cout << endl;
    }*/
}

double ModelIRBSingle::PointToLineDistance(const MSLVector &point,
					   const MSLVector &pointOnLine,
					   const MSLVector &lineVector)
{
  MSLVector pointOnLine2(3);
  pointOnLine2[0] = pointOnLine[0];
  pointOnLine2[1] = pointOnLine[1];
  pointOnLine2[2] = pointOnLine[2];

  MSLVector lineVector2(3);
  lineVector2[0] = lineVector[0];
  lineVector2[1] = lineVector[1];
  lineVector2[2] = lineVector[2];
  MSLVector f = point - pointOnLine2;

  double angle = lineVector2.angle(f);
  double dist = sin(angle) * f.length();
  return dist;
}

void ModelIRBSingle::test()
{
  cout << "test"<< endl;
}
