#include <GL/gl.h>
#include "path.h"

Path::Path(MSLVector &state, Problem *problem, GeomIRB *geom, QWidget *parent) : QGroupBox(parent), state(state), renderPathEnable(true), renderTreeEnable(true), spline(state.dim())
{
  this->problem = problem;

  rrt = new RRTConCon(problem);
  rrt->Holonomic = true;
  rrt->PlannerDeltaT = 0.01;
  rrt->NumNodes = 1000000;

  current = rrt->Path.begin();

  layout = new QFormLayout;
  setLayout(layout);

  QPushButton *calculatePathButton = new QPushButton("Calculate path");
  layout->setWidget(0,QFormLayout::FieldRole, calculatePathButton);
  connect(calculatePathButton, SIGNAL(clicked()),
	  this, SLOT(calculatePath()));

  QPushButton *stepForwardButton = new QPushButton("Step forward");
  layout->setWidget(1,QFormLayout::FieldRole, stepForwardButton);
  connect(stepForwardButton, SIGNAL(clicked()),
	  this, SLOT(stepForward()));

  QPushButton *stepBackwardButton = new QPushButton("Step backward");
  layout->setWidget(2,QFormLayout::FieldRole, stepBackwardButton);
  connect(stepBackwardButton, SIGNAL(clicked()),
	  this, SLOT(stepBackward()));

  QPushButton *renderPathButton = new QPushButton("Render path");
  layout->setWidget(3,QFormLayout::FieldRole, renderPathButton);
  connect(renderPathButton, SIGNAL(clicked()),
	  this, SLOT(renderPath()));

  QPushButton *renderTreeButton = new QPushButton("Render tree");
  layout->setWidget(4, QFormLayout::FieldRole, renderTreeButton);
  connect(renderTreeButton, SIGNAL(clicked()),
	  this, SLOT(renderTree()));

  QPushButton *optimizePathButton = new QPushButton("Remove redundant nodes");
  layout->setWidget(5,QFormLayout::FieldRole, optimizePathButton);
  connect(optimizePathButton, SIGNAL(clicked()),
	  this, SLOT(optimizePath()));

  QPushButton *dividePathButton = new QPushButton("Run Shortcut");
  layout->setWidget(6,QFormLayout::FieldRole, dividePathButton);
  connect(dividePathButton, SIGNAL(clicked()),
	  this, SLOT(dividePath()));

  QPushButton *exportPathButton = new QPushButton("Export path");
  layout->setWidget(7,QFormLayout::FieldRole, exportPathButton);
  connect(exportPathButton, SIGNAL(clicked()),
	  this, SLOT(exportPath()));

  satisfiedButton = new QPushButton("No Collisions");
  layout->setWidget(8,QFormLayout::SpanningRole, satisfiedButton);

  QPushButton *gotoInitialStateButton = new QPushButton("Go to initial state");
  layout->setWidget(9,QFormLayout::FieldRole, gotoInitialStateButton);
  connect(gotoInitialStateButton, SIGNAL(clicked()),
	  this, SLOT(gotoInitialState()));

  QPushButton *gotoFinalStateButton = new QPushButton("Go to final state");
  layout->setWidget(10,QFormLayout::FieldRole, gotoFinalStateButton);
  connect(gotoFinalStateButton, SIGNAL(clicked()),
	  this, SLOT(gotoFinalState()));

  QPushButton *storePathButton = new QPushButton("Store path");
  layout->setWidget(11,QFormLayout::FieldRole, storePathButton);
  connect(storePathButton, SIGNAL(clicked()),
	  this, SLOT(storePath()));

  QPushButton *restorePathButton = new QPushButton("Restore path");
  layout->setWidget(12,QFormLayout::FieldRole, restorePathButton);
  connect(restorePathButton, SIGNAL(clicked()),
	  this, SLOT(restorePath()));

  
  deltaT = 0.01f;

  splineRenderer = new SplineRenderer(&spline, problem);
  treeRenderer = new TreeRenderer(rrt, problem);

  objective = new PathOptimization;
  objective->setSpline(&spline);
  objective->setProblem(problem);
  objective->setGeom(geom);
}

void Path::setRobot(ModelIRBSingle *irb)
{
  objective->setRobot(irb);
}

PathOptimization* Path::getObjective()
{
  return objective;
}

void Path::calculatePath()
{
  if (rrt->Plan())
    cout << "Solution found" << endl;

  current = rrt->Path.begin();

  spline.clear();
  t = 0.0f;

  for (list <MSLVector>::iterator it = rrt->Path.begin(); it != rrt->Path.end(); it++) {
    spline.add(*it);
  }
}

void Path::stepForward()
{
  if (rrt->Path.size() == 0)
    return;

  if (t+deltaT <= 1.0)
    t += deltaT;

  state = spline.evaluate(t);
  if (problem->Satisfied(state)) {
    satisfiedButton->setText("No Collisions");
  } else {
    satisfiedButton->setText("Collision");
  }
  emit stateChanged();
}

void Path::stepBackward()
{
  if (rrt->Path.size() == 0)
    return;

  if (t-deltaT >= 0.0f)
    t -= deltaT;

  state = spline.evaluate(t);
  if (problem->Satisfied(state)) {
    satisfiedButton->setText("No Collisions");
  } else {
    satisfiedButton->setText("Collision");
  }
  emit stateChanged();
}

void Path::gotoInitialState()
{
  state = problem->InitialState;
  emit stateChanged();
}

void Path::gotoFinalState()
{
  state = problem->GoalState;
  emit stateChanged();
}

void Path::renderPath()
{
  renderPathEnable = ! renderPathEnable;
}

void Path::renderTree()
{
  renderTreeEnable = ! renderTreeEnable;
}

void Path::optimizePath()
{
  //  for (unsigned int i=0; i<5; i++) {
    objective->straighten();
    emit stateChanged();
    //  }
}

void Path::dividePath()
{
  //  for (int i=0; i<30; i++)
    objective->divide();
  emit stateChanged();
}


void Path::exportPath()
{
  cout << "Point list start" << endl;
  for (unsigned int i=0; i<spline.numPoints(); i++) {
    cout << spline[i] << endl;
  }
  cout << "Point list end" << endl;
}

void Path::storePath()
{
  spline.store();
}

void Path::restorePath()
{
  spline.restore();
}

void Path::render() {
  glLineWidth(3.0);
  glDisable(GL_LIGHTING);
  if (renderPathEnable) {
    glColor3f(0.0, 1.0, 1.0);
    splineRenderer->render();
  }
  if (renderTreeEnable) {
    glColor3f(1.0, 0.0, 0.0);
    treeRenderer->render();
  }
  glEnable(GL_LIGHTING);
}

#include "moc_path.cpp"
