#include <cstdlib>
#include <ctime>

#include "pathoptimization.h"

PathOptimization::PathOptimization()
{
  spline  = NULL;
  problem = NULL;
  irb     = NULL;
}

void PathOptimization::setSpline(Spline *spline)
{
  this->spline = spline;
}

void PathOptimization::setProblem(Problem *problem)
{
  this->problem = problem;
}

void PathOptimization::setGeom(GeomIRB* geom)
{
  this->geom = geom;
}

void PathOptimization::setRobot(ModelIRBSingle *irb)
{
  this->irb = irb;
}

bool PathOptimization::lineCheck(Vector &q1, Vector &q2, unsigned int depth = 1)
{
  Vector q1Comp = problem->StateToConfiguration(q1);
  GeomIRB_DistanceResult dist = geom->DistanceCompute(q1Comp);
  
  Vector deltaq = q2-q1;

  Vector conf = problem->StateToConfiguration(q1);
  if (!problem->CollisionFree(conf)) {
    return false;
  }
  conf = problem->StateToConfiguration(q2);
  if (!problem->CollisionFree(conf)) {
    return false;
  }
  
  double sum;

  float c[36];

  irb->LipschitzConstants(q1, c);

  bool test = true;

  for (int i=0; i<6; i++) {
    sum = 0.0f;
    for (int j=5; j>=0; j--) {
      sum += (sum + c[i+j*6]) * fabs(deltaq[j]);
    }
    if (sum >= dist.distance[i]) {
      test = false;
      break;
    }
  }

  Vector midpoint = (q1+q2)*0.5;

  if (test == false && depth != 0) {
    return lineCheck(q1, midpoint, depth--) && lineCheck(midpoint, q2, depth--);
  }

  return test;
}

void PathOptimization::straighten()
{
  std::vector<int> positionsToRemove;

  unsigned int numberOfPointsRemoved = 0;
  bool didRemoveSomething = true;
  while(didRemoveSomething) {
    didRemoveSomething = false;
    for (unsigned int i=1; i<spline->numPoints()-2; i+=2) {
      Vector &a = (*spline)[i-1];
      Vector &c = (*spline)[i+1];
      if (lineCheck(a,c,10)) {
	positionsToRemove.push_back(i-positionsToRemove.size());
	i++;
	numberOfPointsRemoved ++;
	didRemoveSomething = true;
      }
    }
    for (unsigned int i=0; i<positionsToRemove.size(); i++) {
      spline->remove(positionsToRemove[i]);
    }
    positionsToRemove.clear();
  }



  cout << "Removed " << numberOfPointsRemoved << " points. There are " << spline->numPoints() << " left."<< endl;
}

void PathOptimization::shortcut()
{
  srand(time(NULL));
  Vector a(6);
  Vector b(6);
  double num1, num2;
  double mini, maxi;

  unsigned int numberOfShortcuts = 0;
  
  for (unsigned int i=1; i<100; i++) {
    num1 = rand()/(double)RAND_MAX;
    num2 = rand()/(double)RAND_MAX;

    mini = min(num1,num2);
    maxi = max(num1,num2);


    a = spline->evaluate(mini);
    b = spline->evaluate(maxi);
   

    if (lineCheck(a,b,10)) {
      spline->shortcut(mini, maxi);
      numberOfShortcuts ++;
    }
  }
  cout << "Number of shortcuts: " << numberOfShortcuts<<" There are "<<spline->numPoints() << " points" << endl;
}

void PathOptimization::straighten2()
{  

  for (unsigned int i=1; i<spline->numPoints()-1; i++) {
    Vector &a = (*spline)[i-1];
    Vector &b = (*spline)[i];
    Vector &c = (*spline)[i+1];

    Vector d1 = a - b;
    Vector d2 = c - b;
    Vector d3 = (d1+d2)*0.1;

    unsigned int depth = 3;
    Vector tempState1 = (a+c)*0.5;
    Vector tempState2 = b + d3;
    if (lineCheck(a, tempState1, depth) && lineCheck(tempState1, c, depth)) {
      cout << "Move halfway" << endl;
      b = tempState1;
    } else if (lineCheck(a, tempState2,depth) && lineCheck(tempState2, c,depth)) {
      cout << "Move a distance" << endl;
      b = tempState2;
    }

  }
}

void PathOptimization::divide()
{
  shortcut();
}

