#include <QtGui>
#include "window.h"
#include "link.h"
#include "transform.h"
#include "path.h"
#include "controls.h"

Window::Window(QWidget *parent)  : QWidget(parent)
{  
  irbrenderer = NULL;
  path = NULL;
  environment = NULL;
  resize(1900, 1000);

  QHBoxLayout *main = new QHBoxLayout;
  setLayout(main);

  glWidget = new GLWidget();
  main->addWidget(glWidget,3);
  connect(glWidget, SIGNAL(mustUpdate()),
	  this, SLOT(refresh()));

  tabs = new QTabWidget();
  main->addWidget(tabs,1);

  controls = new Controls(glWidget->getCamera());
  tabs->addTab(controls, "Controls");

  settings = new Settings;
  tabs->addTab(settings, "Settings");
  connect(settings, SIGNAL(update()),
	  this, SLOT(refresh()));

  environment = NULL;
  /*
  environment = new Environment(p+string("/Obst"));
  environment->setColor(1.0, 1.0, 0.0);

  Transform *eTransform = new Transform(string("Environment"), environment);
  tabs->addTab(eTransform, string("Environment").c_str());
  connect(eTransform, SIGNAL(update()),
  this, SLOT(refresh()));*/


  updateConfiguration();

  update();
}

void Window::update()
{
  glWidget->repaint();
  if (irbrenderer != NULL) {
    if (settings->showGraphicalModel()) {
      irbrenderer->renderGraphicalModels();
    }
    if (settings->showCollisionModel()) {
      irbrenderer->renderCollisionModels();
    }
    if (settings->showCoordinateSystems()) {
      irbrenderer->renderCoordinateSystems();
    }
    if (settings->showLineModel()) {
      irbrenderer->renderLineModel();
    }
  }
  if (path != NULL) {
    path->render();
  }

  if (settings->showEnvironment()) {
    if (environment != NULL) {
      environment->draw();
    }
  }

  glWidget->swapBuffers();
}

void Window::setEnvironment(Environment *environment)
{
  this->environment = environment;
  //  environment = new Environment(p+string("/Obst"));
  environment->setColor(1.0, 1.0, 0.0);

  Transform *eTransform = new Transform(string("Environment"), environment);
  tabs->addTab(eTransform, string("Environment").c_str());
  connect(eTransform, SIGNAL(update()),
	  this, SLOT(refresh()));

}

void Window::setModel(ModelIRB *_model, string _path)
{
  model = _model;

  configuration = MSLVector(model->NumBodies);

  geom = new GeomIRB(_path);
  prob = new Problem(geom, model, _path);

  for (int i=0; i<model->NumBodies; i++) {
    controls->addLink(model->LowerState[i], model->UpperState[i]);
  }


  irbrenderer = new IRBRenderer(_path, model);
  for (unsigned int i=0; i<irbrenderer->numLinks(); i++) {
    tabs->addTab(irbrenderer->getControlWidget(i), irbrenderer->getName(i).c_str());
  }
  connect(irbrenderer, SIGNAL(update()),
	  this, SLOT(refresh()));



  path = new Path(configuration, prob, geom);
  tabs->insertTab(2,path, "Path");

  connect(path, SIGNAL(stateChanged()),
	  this, SLOT(stateToConfiguration()));
  connect(controls, SIGNAL(cameraChanged()),
	  this, SLOT(updateCamera()));
  connect(controls, SIGNAL(stateChanged()),
	  this, SLOT(updateConfiguration()));
  connect(controls, SIGNAL(printConfiguration()),
	  this, SLOT(printConfiguration()));

  updateConfiguration();
}

PathOptimization* Window::getObjective()
{
  return path->getObjective();
}

void Window::refresh()
{
  update();
}

void Window::updateCamera()
{
  update();
}

void Window::stateToConfiguration()
{
  if (configuration.dim() > 0) {
    irbrenderer->stateToConfiguration(configuration);
    controls->setState(configuration);
    MSLVector q = model->StateToConfiguration(configuration);
    //  cout << "eps: " << fabs(model->dist(q)-model->distance) << endl;
  }
}

void Window::updateConfiguration()
{
  configuration = controls->getState();
  stateToConfiguration();
}

void Window::printConfiguration()
{
  cout << "Configuration: " << endl;
  for (int i=0; i<configuration.dim(); i++) {
    cout << configuration[i] << endl;
  }
}


#include "moc_window.cpp"
